// SPDX-FileCopyrightText: 2019-2024 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "DrawOverviewPage.h"
#include "DrawModel.h"
#include "DrawView.h"

#include "SharedObjects/SharedObjects.h"
#include "SharedObjects/SearchEngine.h"
#include "SharedObjects/TournamentSettings.h"

#include "shared/TableDelegate.h"
#include "shared/SearchWidget.h"
#include "shared/DrawMapper.h"

#include "RegistrationPage/RegistrationsModel.h"

// Qt includes
#include <QGroupBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QPushButton>
#include <QApplication>
#include <QStyle>
#include <QLabel>
#include <QDebug>

DrawOverviewPage::DrawOverviewPage(QWidget *parent, SharedObjects *sharedObjects,
                                   RegistrationsModel *registrationsModel,
                                   bool drawAdjustmentNeeded)
    : OptionalPage(parent),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_searchEngine(sharedObjects->searchEngine()),
      m_registrationsModel(registrationsModel)
{
    // Top layout

    auto *topLayout = new QHBoxLayout;
    layout()->addLayout(topLayout);

    auto *statusLayout = new QVBoxLayout;
    topLayout->addLayout(statusLayout);

    // Status labels

    m_uncompletedTablesLabel = new QLabel;
    m_uncompletedTablesLabel->setWordWrap(true);
    statusLayout->addWidget(m_uncompletedTablesLabel);

    m_adjustmentLabel = new QLabel(tr("Alle %1 haben eine Auslosung, aber die Vergabe ist nicht "
                                      "fortlaufend. <a href=\"#adjustDraw\">Auslosung bereinigen"
                                      "</a>").arg(
                                      m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                         : tr("Spieler")));
    m_adjustmentLabel->setWordWrap(true);
    m_adjustmentLabel->setTextInteractionFlags(Qt::TextBrowserInteraction);
    connect(m_adjustmentLabel, &QLabel::linkActivated,
            this, &DrawOverviewPage::showAdjustDrawDialog);
    m_adjustmentLabel->setVisible(drawAdjustmentNeeded);
    statusLayout->addWidget(m_adjustmentLabel);

    // Close button
    auto *closeButton = new QPushButton(tr("Schließen"));
    closeButton->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    closeButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCloseButton));
    connect(closeButton, &QPushButton::clicked, this, &DrawOverviewPage::closePage);
    topLayout->addWidget(closeButton);

    // Draw overview

    auto *box = new QGroupBox;
    auto *boxLayout = new QVBoxLayout(box);
    layout()->addWidget(box);

    m_noDraw = new QLabel(tr("<i>Es wurde noch keine Auslosung eingegeben</i>"));
    m_noDraw->setAlignment(Qt::AlignCenter);
    boxLayout->addWidget(m_noDraw);

    m_isSearch = new QLabel;
    m_isSearch->hide();
    boxLayout->addWidget(m_isSearch);

    m_drawModel = new DrawModel(this, &m_registrationsModel->playersData(), sharedObjects);
    m_drawView = new DrawView;
    m_drawView->setModel(m_drawModel);
    boxLayout->addWidget(m_drawView);

    // Search widget

    m_searchWidget = new SearchWidget(QStringLiteral("draw_page_search"), sharedObjects);
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &DrawOverviewPage::updateSearch);
    boxLayout->addWidget(m_searchWidget);
}

bool DrawOverviewPage::closePage()
{
    if (m_tournamentSettings->isSinglePlayerMode()) {
        m_searchEngine->clearSearchCache(SearchEngine::DrawnPairsDrawSearch);
    }

    return OptionalPage::closePage();
}

void DrawOverviewPage::updateDraw(const QString &scrollTarget)
{
    m_drawModel->refresh(m_tournamentSettings->isPairMode());

    if (m_drawModel->rowCount() == 0) {
        m_noDraw->show();
        m_uncompletedTablesLabel->setText(QStringLiteral(" ")); // Act as a placeholder
        m_drawView->hide();
        m_searchWidget->clear();
        m_searchWidget->hide();
        m_adjustmentLabel->hide();
        return;
    } else {
        m_noDraw->hide();
        m_drawView->show();
        m_searchWidget->show();
    }

    QString text = QStringLiteral("<table>");

    const auto &incompleteTables = m_registrationsModel->incompleteTables();

    const int incompleteTablesCount = incompleteTables.count();
    if (incompleteTablesCount > 0) {
        text.append(tr("<tr><td>Unvollständig besetzte Tische:</td><td>"));
        for (int i = 0; i < incompleteTablesCount; i++) {
            if (i > 0 && incompleteTablesCount > 1) {
                if (i < incompleteTablesCount - 1) {
                    text.append(tr(", "));
                } else {
                    text.append(tr(" und "));
                }
            }
            text.append(QString::number(incompleteTables.at(i)));
        }
        text.append(QStringLiteral("</td></tr>"));
    }

    const auto &tableGaps = m_registrationsModel->tableGaps();
    const int tableGapsCount = tableGaps.count();
    if (tableGapsCount > 0) {
        text.append(tr("<tr><td>Unbesetzte Tische:</td><td>"));
        for (int i = 0; i < tableGapsCount; i++) {
            if (i > 0 && tableGapsCount > 1) {
                if (i < tableGapsCount - 1) {
                    text.append(tr(", "));
                } else {
                    text.append(tr(" und "));
                }
            }
            text.append(QString::number(tableGaps.at(i)));
        }
        text.append(QStringLiteral("</td></tr>"));
    }

    if (text == QStringLiteral("<table>")) {
        m_uncompletedTablesLabel->setText(tr("Die Auslosung ist vollständig und lückenlos"));
    } else {
        text.append(QStringLiteral("</table>"));
        m_uncompletedTablesLabel->setText(text);
    }

    if (! scrollTarget.isEmpty()) {
        m_drawView->scrollToName(scrollTarget);
    }
}

void DrawOverviewPage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_drawModel->rowCount(); row++) {
            m_drawView->showRow(row);
        }
        m_drawView->resizeColumnsToContents();
        m_isSearch->hide();
        return;
    }

    int matches = 0;

    bool isNumber = false;
    const auto number = searchText.toInt(&isNumber);

    if (isNumber && number > 0) {
        // We do a pure number search and check for the players number only
        for (int row = 0; row < m_drawModel->rowCount(); row += 2) {
            if (m_drawModel->tableMatchesNumber(row, number)) {
                matches++; // This can actually only happen once
                m_drawView->showRow(row);
                m_drawView->showRow(row + 1);
            } else {
                m_drawView->hideRow(row);
                m_drawView->hideRow(row + 1);
            }
        }

    } else {
        // We query the search engine
        m_searchEngine->setSearchTerm(searchText);
        const auto phoneticSearch = m_searchWidget->phoneticSearch();

        for (int row = 0; row < m_drawModel->rowCount(); row += 2) {
            if (m_drawModel->tableMatchesSearchTerm(row, phoneticSearch)) {
                matches++;
                m_drawView->showRow(row);
                m_drawView->showRow(row + 1);
            } else {
                m_drawView->hideRow(row);
                m_drawView->hideRow(row + 1);
            }
        }
    }

    m_searchWidget->updateMatches(matches);
    m_drawView->resizeColumnsToContents();
    m_isSearch->setText(tr("Suchergebnis für „%1“").arg(searchText.simplified()));
    m_isSearch->show();
}

void DrawOverviewPage::resetSearch()
{
    m_searchWidget->clear();
}

void DrawOverviewPage::drawAdjustmentNeeded(bool state)
{
    m_adjustmentLabel->setVisible(state);
}
