// SPDX-FileCopyrightText: 2020-2024 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "BasicSettingsPage.h"

#include "SharedObjects/TournamentSettings.h"

#include "shared/TournamentTemplate.h"
#include "shared/DefaultValues.h"

// Qt includes
#include <QLabel>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QRadioButton>
#include <QSpinBox>

BasicSettingsPage::BasicSettingsPage(const TournamentTemplate::Settings &tournamentTemplate,
                                     TournamentSettings *tournamentSettings,
                                     QWidget *parent)
    : AbstractSettingsPage(tr("Grundeinstellungen"), parent),
      m_tournamentSettings(tournamentSettings)
{
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);

    QLabel *tournamentSettingsLabel = new QLabel(tr(
        "Hier werden die Grundparameter des Muckturniers festgelegt. Diese Einstellungen sind "
        "fix und können während des Turniers nicht mehr verändert werden."));
    tournamentSettingsLabel->setWordWrap(true);
    layout()->addWidget(tournamentSettingsLabel);

    // Tournament type

    QGroupBox *tournamentTypeBox = new QGroupBox(tr("Turniertyp"));
    QGridLayout *tournamentTypeBoxLayout = new QGridLayout(tournamentTypeBox);
    layout()->addWidget(tournamentTypeBox);

    m_fixedPairs = new QRadioButton(tr("Feste Paare"));
    m_fixedPairs->setChecked(tournamentTemplate.tournamentMode == Tournament::FixedPairs);
    connect(m_fixedPairs, &QRadioButton::toggled,
            this, [this](bool checked)
            {
                Q_EMIT tournamentModeSelected(checked ? Tournament::FixedPairs
                                                      : Tournament::SinglePlayers);
            });
    tournamentTypeBoxLayout->addWidget(m_fixedPairs, 0, 0, Qt::AlignTop);

    tournamentTypeBoxLayout->addWidget(
        descriptionLabel(tr("Es werden immer zwei Spieler als festes Paar angemeldet, das das "
                            "ganze Turnier zusammen spielt.")),
        0, 1);

    m_singlePlayers = new QRadioButton(tr("Einzelne Spieler"));
    m_singlePlayers->setChecked(! m_fixedPairs->isChecked());
    tournamentTypeBoxLayout->addWidget(m_singlePlayers, 1, 0, Qt::AlignTop);

    tournamentTypeBoxLayout->addWidget(
        descriptionLabel(tr("Es werden einzelne Spieler angemeldet.\n"
                            "Die Paare werden (ggf. pro Runde) ausgelost.")),
        1, 1);

    // Boogers and scores

    QGroupBox *boogersBox = new QGroupBox(tr("Bobbl und Punkte"));
    QHBoxLayout *boogerWrapperLayout = new QHBoxLayout(boogersBox);
    QGridLayout *boogerLayout = new QGridLayout;
    boogerWrapperLayout->addLayout(boogerLayout);
    layout()->addWidget(boogersBox);

    boogerLayout->addWidget(new QLabel(tr("Bobbl pro Runde:")), 1, 0);
    m_boogersPerRound = new QSpinBox;
    m_boogersPerRound->setMinimum(1);
    m_boogersPerRound->setMaximum(DefaultValues::maximumBoogersPerRound);
    m_boogersPerRound->setValue(tournamentTemplate.boogersPerRound);
    boogerLayout->addWidget(m_boogersPerRound, 1, 1);

    boogerLayout->addWidget(new QLabel(tr("Punkte pro Bobbl:")), 0, 0);
    m_boogerScore = new QSpinBox;
    m_boogerScore->setMinimum(1);
    m_boogerScore->setMaximum(DefaultValues::maximumBoogerScore);
    m_boogerScore->setValue(tournamentTemplate.boogerScore);
    boogerLayout->addWidget(m_boogerScore, 0, 1);

    boogerWrapperLayout->addStretch();

    layout()->addStretch();

    m_advancedOptionsLabel = new QLabel(tr(
        "Sämtliche weitere Optionen können während des Turniers geändert werden. Hierfür kann eine "
        "Vorlage erstellt werden, so dass alle Einstellungen gleich zu Beginn passen. Für alle "
        "Einstellungen „Erweiterte Optionen“ anhaken!"
    ));
    m_advancedOptionsLabel->setWordWrap(true);
    layout()->addWidget(m_advancedOptionsLabel);

    layout()->addStretch();
}

void BasicSettingsPage::extendedOptionsToggled(bool state)
{
    m_advancedOptionsLabel->setVisible(! state);
}

void BasicSettingsPage::saveSettings()
{
    m_tournamentSettings->setTournamentMode(m_fixedPairs->isChecked() ? Tournament::FixedPairs
                                                                      : Tournament::SinglePlayers);
    m_tournamentSettings->setBoogerScore(m_boogerScore->value());
    m_tournamentSettings->setBoogersPerRound(m_boogersPerRound->value());
}
