// SPDX-FileCopyrightText: 2019-2024 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "TournamentSettings.h"

#include "shared/JsonHelper.h"
#include "shared/TournamentTemplate.h"
#include "shared/Markers.h"
#include "shared/Json.h"
#include "shared/Logging.h"

// Qt includes
#include <QDebug>
#include <QJsonObject>
#include <QJsonArray>

// C++ includes
#include <utility>

namespace tt = TournamentTemplate;

// The revision we use
static constexpr int s_tournamentSettingsVersion = 2;

// JSON Keys for all settings

static const QLatin1String s_muckturnierVersion("muckturnier version");
static const QLatin1String s_tournamentMode    ("tournament mode");
static const QLatin1String s_boogerScore       ("booger score");
static const QLatin1String s_boogersPerRound   ("boogers per round");

static const QLatin1String s_tournamentOpen("tournament open");

static const QLatin1String s_scoreType              ("score type");
static const QLatin1String s_noTableNumbers         ("no table numbers");
static const QLatin1String s_autoSelectPairs        ("autoselect pairs");
static const QLatin1String s_selectByLastRound      ("select by last round");
static const QLatin1String s_includeOpponentGoals   ("include opponent goals");
static const QLatin1String s_grayOutUnimportantGoals("gray out unimportant goals");

static const QLatin1String s_defaultMarker           ("default marker");
static const QLatin1String s_drawnMarker             ("drawn marker");
static const QLatin1String s_absentMarker            ("absent marker");
static const QLatin1String s_singlesManagementVersion("singles management version");
static const QLatin1String s_singlesManagement       ("singles management");

static const QLatin1String s_registrationColumns         ("registration columns");
static const QLatin1String s_registrationListHeaderHidden("registration list header hidden");

static const QLatin1String s_drawSettingsVersion("draw settings version");
static const QLatin1String s_drawSettings       ("draw settings");

static const QLatin1String s_scheduleSettingsVersion("schedule settings version");
static const QLatin1String s_scheduleSettings       ("schedule settings");

static const QLatin1String s_bookingSettings("booking settings");

static const QLatin1String s_optionalPages("optional pages");

TournamentSettings::TournamentSettings(QObject *parent) : QObject(parent)
{
    // Initialize all variables to default values
    reset();
}

int TournamentSettings::version() const
{
    return s_tournamentSettingsVersion;
}

bool TournamentSettings::loadingDb() const
{
    return m_loadingDb;
}

void TournamentSettings::setLoadingDb(bool state)
{
    m_loadingDb = state;
}

const QVector<OptionalPages::Page> &TournamentSettings::optionalPages() const
{
    return m_optionalPages;
}

void TournamentSettings::setOptionalPageShown(OptionalPages::Page page, bool state)
{
    if (m_loadingDb) {
        return;
    }

    if (state) {
        m_optionalPages.append(page);
    } else {
        m_optionalPages.removeOne(page);
    }
}

void TournamentSettings::setDbTournamentSettingsVersion(int version)
{
    m_dbTournamentSettingsVersion = version;
}

int TournamentSettings::dbTournamentSettingsVersion() const
{
    return m_dbTournamentSettingsVersion;
}

void TournamentSettings::reset()
{
    // Initialize the default values from the tournament template
    m_tournamentMode               = tt::defaultSettings.tournamentMode;
    m_scoreType                    = tt::defaultSettings.scoreType;
    m_boogerScore                  = tt::defaultSettings.boogerScore;
    m_boogersPerRound              = tt::defaultSettings.boogersPerRound;
    m_drawSettings                 = tt::defaultSettings.drawSettings;
    m_scheduleSettings             = tt::defaultSettings.scheduleSettings;
    m_registrationColumns          = tt::defaultSettings.registrationColumns;
    m_registrationListHeaderHidden = tt::defaultSettings.registrationListHeaderHidden;
    m_autoSelectPairs              = tt::defaultSettings.autoSelectPairs;
    m_selectByLastRound            = tt::defaultSettings.selectByLastRound;
    m_noTableNumbers               = tt::defaultSettings.noTableNumbers;
    m_includeOpponentGoals         = tt::defaultSettings.includeOpponentGoals;
    m_grayOutUnimportantGoals      = tt::defaultSettings.grayOutUnimportantGoals;
    m_optionalPages                = tt::defaultSettings.optionalPages;

    // Fallback values for database-read settings
    m_muckturnierVersion     = tr("unbekannt");
    m_tournamentOpen         = true;
    m_originalTournamentOpen = true;
    m_marker                 = Markers::defaultSpecialMarkerValues;
    m_singlesManagement      = Markers::defaultSinglesManagement;
    m_bookingSettings        = QString();
}

QVariant TournamentSettings::getValue(const QJsonObject &json, const QLatin1String &key,
                                      QJsonValue::Type type) const
{
    const auto value = json.value(key);
    if (value.isUndefined() || value.type() != type) {
        return QVariant();
    }
    return value.toVariant();
}

bool TournamentSettings::load(const QString &data)
{
    // Set default fall-back values
    reset();

    // Clear cached unhandled values
    m_unhandledValues.clear();

    // Parse the JSON data
    const auto json = Json::objectFromString(data);
    QVariant value;

    // There are some essential values that have to be present and valid. We parse them here:

    // tournamentMode

    value = getValue(json, s_tournamentMode, QJsonValue::String);
    if (! value.isValid()) {
        return false;
    }

    const auto tournamentMode = value.toString();
    bool tournamentModeValid = false;
    QHash<Tournament::Mode, QString>::const_iterator it;
    for (it = Tournament::modeMap.constBegin(); it != Tournament::modeMap.constEnd(); it++) {
        if (it.value() == tournamentMode) {
            m_tournamentMode = it.key();
            tournamentModeValid = true;
            break;
        }
    }
    if (! tournamentModeValid) {
        return false;
    }

    // boogerScore
    value = getValue(json, s_boogerScore, QJsonValue::Double);
    if (! value.isValid()) {
        return false;
    }
    m_boogerScore = value.toInt();

    // boogersPerRound
    value = getValue(json, s_boogersPerRound, QJsonValue::Double);
    if (! value.isValid()) {
        return false;
    }
    m_boogersPerRound = value.toInt();

    // For all other values, we can fall back to the default setting set by reset()
    // so we only set them if they are present or actually parsed

    // scoreType

    value = getValue(json, s_scoreType, QJsonValue::String);
    const auto scoreType = value.toString();

    QHash<Tournament::ScoreType, QString>::const_iterator it2;
    for (it2 = Tournament::scoreTypeMap.constBegin(); it2 != Tournament::scoreTypeMap.constEnd();
         it2++) {

        if (it2.value() == scoreType) {
            m_scoreType = it2.key();
            break;
        }
    }

    // drawSettings
    const bool drawSettingsParseable = json.value(s_drawSettingsVersion).toInt()
                                           == Draw::settingsVersion;
    if (drawSettingsParseable) {
        m_drawSettings = JsonHelper::parseDrawSettings(json.value(s_drawSettings).toObject());
    }

    // scheduleSettings
    const auto scheduleSettingsVersion = json.value(s_scheduleSettingsVersion).toInt();
    const auto scheduleSettingsParseable = scheduleSettingsVersion == Schedule::settingsVersion;
    if (scheduleSettingsParseable) {
        m_scheduleSettings = JsonHelper::parseScheduleSettings(
            json.value(s_scheduleSettings).toObject());
    } else {
        qCWarning(MuckturnierLog) << "Schedule settings with version " <<  scheduleSettingsVersion
                                  << "are not parseable, falling back to defaults";
    }

    // registrationColumns
    m_registrationColumns = JsonHelper::parseRegistrationColumns(
                                json.value(s_registrationColumns).toArray());

    // registrationListHeaderHidden
    value = getValue(json, s_registrationListHeaderHidden, QJsonValue::Bool);
    if (value.isValid()) {
        m_registrationListHeaderHidden = value.toBool();
    }

    // autoSelectPairs
    value = getValue(json, s_autoSelectPairs, QJsonValue::Bool);
    if (value.isValid()) {
        m_autoSelectPairs = value.toBool();
    }

    // selectByLastRound
    value = getValue(json, s_selectByLastRound, QJsonValue::Bool);
    if (value.isValid()) {
        m_selectByLastRound = value.toBool();
    }

    // noTableNumbers
    value = getValue(json, s_noTableNumbers, QJsonValue::Bool);
    if (value.isValid()) {
        m_noTableNumbers = value.toBool();
    }

    // includeOpponentGoals
    value = getValue(json, s_includeOpponentGoals, QJsonValue::Bool);
    if (value.isValid()) {
        m_includeOpponentGoals = value.toBool();
    }

    // grayOutUnimportantGoals
    value = getValue(json, s_grayOutUnimportantGoals, QJsonValue::Bool);
    if (value.isValid()) {
        m_grayOutUnimportantGoals = value.toBool();
    }

    // optionalPages
    m_optionalPages = JsonHelper::parseOptionalPages(json.value(s_optionalPages).toArray());

    // muckturnierVersion
    value = getValue(json, s_muckturnierVersion, QJsonValue::String);
    if (value.isValid()) {
        m_muckturnierVersion = value.toString();
    }

    // tournamentOpen
    value = getValue(json, s_tournamentOpen, QJsonValue::Bool);
    if (value.isValid()) {
        m_tournamentOpen = value.toBool();
        m_originalTournamentOpen = m_tournamentOpen;
    }

    // Special markers

    value = getValue(json, s_defaultMarker, QJsonValue::Double);
    if (value.isValid()) {
        m_marker[Markers::Default] = value.toInt();
    }

    value = getValue(json, s_drawnMarker, QJsonValue::Double);
    if (value.isValid()) {
        m_marker[Markers::Drawn] = value.toInt();
    }

    value = getValue(json, s_absentMarker, QJsonValue::Double);
    if (value.isValid()) {
        m_marker[Markers::Absent] = value.toInt();
    }

    // singlesManagement
    const bool singlesManagementParseable = json.value(s_singlesManagementVersion).toInt()
                                                == Markers::singlesManagementVersion;
    if(singlesManagementParseable) {
        m_singlesManagement
            = JsonHelper::parseSinglesManagement(json.value(s_singlesManagement).toObject());
    }

    // Booking settings
    value = getValue(json, s_bookingSettings, QJsonValue::String);
    if (value.isValid()) {
        m_bookingSettings = value.toString();
    }

    // Check for unhandled values

    if (m_dbTournamentSettingsVersion > s_tournamentSettingsVersion) {
        // The settings were created by a newer version of Muckturnier. Some keys are not handled by
        // this version yet. To keep them when closing and saving the currrent settings, we cache
        // them here so that they can be re-added later.

        auto unhandledKeys = json.keys();
        unhandledKeys.removeOne(s_muckturnierVersion);
        unhandledKeys.removeOne(s_tournamentMode);
        unhandledKeys.removeOne(s_tournamentOpen);
        unhandledKeys.removeOne(s_scoreType);
        unhandledKeys.removeOne(s_boogerScore);
        unhandledKeys.removeOne(s_boogersPerRound);
        unhandledKeys.removeOne(s_noTableNumbers);
        unhandledKeys.removeOne(s_autoSelectPairs);
        unhandledKeys.removeOne(s_selectByLastRound);
        unhandledKeys.removeOne(s_includeOpponentGoals);
        unhandledKeys.removeOne(s_grayOutUnimportantGoals);
        unhandledKeys.removeOne(s_defaultMarker);
        unhandledKeys.removeOne(s_drawnMarker);
        unhandledKeys.removeOne(s_absentMarker);
        if (singlesManagementParseable) {
            unhandledKeys.removeOne(s_singlesManagementVersion);
            unhandledKeys.removeOne(s_singlesManagement);
        }
        unhandledKeys.removeOne(s_registrationColumns);
        unhandledKeys.removeOne(s_registrationListHeaderHidden);
        if (drawSettingsParseable) {
            unhandledKeys.removeOne(s_drawSettingsVersion);
            unhandledKeys.removeOne(s_drawSettings);
        }
        if (scheduleSettingsParseable) {
            unhandledKeys.removeOne(s_scheduleSettingsVersion);
            unhandledKeys.removeOne(s_scheduleSettings);
        }
        unhandledKeys.removeOne(s_optionalPages);
        unhandledKeys.removeOne(s_bookingSettings);

        for (const QString &key : std::as_const(unhandledKeys)) {
            m_unhandledValues.insert(key, json.value(key));
        }
    }

    return true;
}

QString TournamentSettings::toString() const
{
    QJsonObject settings {
        { s_muckturnierVersion,           m_muckturnierVersion },
        { s_tournamentMode,               Tournament::modeMap.value(m_tournamentMode) },
        { s_tournamentOpen,               m_tournamentOpen },
        { s_scoreType,                    Tournament::scoreTypeMap.value(m_scoreType) },
        { s_boogerScore,                  m_boogerScore },
        { s_boogersPerRound,              m_boogersPerRound },
        { s_noTableNumbers,               m_noTableNumbers },
        { s_autoSelectPairs,              m_autoSelectPairs },
        { s_selectByLastRound,            m_selectByLastRound },
        { s_includeOpponentGoals,         m_includeOpponentGoals },
        { s_grayOutUnimportantGoals,      m_grayOutUnimportantGoals },
        { s_defaultMarker,                m_marker.value(Markers::Default) },
        { s_drawnMarker,                  m_marker.value(Markers::Drawn) },
        { s_absentMarker,                 m_marker.value(Markers::Absent) },
        { s_registrationListHeaderHidden, m_registrationListHeaderHidden },

        // Registration columns
        { s_registrationColumns, JsonHelper::registrationColumnsToJson(m_registrationColumns) },

        // Optional pages
        { s_optionalPages, JsonHelper::optionalPagesToJson(m_optionalPages) },

        // Singles management
        { s_singlesManagementVersion, Markers::singlesManagementVersion },
        { s_singlesManagement,        JsonHelper::singlesManagementToJson( m_singlesManagement) },

        // Draw settings
        { s_drawSettingsVersion, Draw::settingsVersion },
        { s_drawSettings,        JsonHelper::drawSettingsToJson(m_drawSettings) },

        // Schedule settings
        { s_scheduleSettingsVersion, Schedule::settingsVersion },
        { s_scheduleSettings,        JsonHelper::scheduleSettingsToJson(m_scheduleSettings) },

        // Booking settings
        { s_bookingSettings, m_bookingSettings }
    };

    // If we have unhandled values, add them as well so that they aren't lost
    if (! m_unhandledValues.isEmpty()) {
        QHash<QString, QJsonValue>::const_iterator it;
        for (it = m_unhandledValues.constBegin(); it != m_unhandledValues.constEnd(); it++) {
            settings.insert(it.key(), it.value());
        }
    }

    return Json::serialize(settings);
}

void TournamentSettings::setMuckturnierVersion(const QString &text)
{
    m_muckturnierVersion = text;
}

const QString &TournamentSettings::muckturnierVersion() const
{
    return m_muckturnierVersion;
}

void TournamentSettings::setTournamentMode(Tournament::Mode tournamentMode)
{
    m_tournamentMode = tournamentMode;
}

Tournament::Mode TournamentSettings::tournamentMode() const
{
    return m_tournamentMode;
}

bool TournamentSettings::isPairMode() const
{
    return m_tournamentMode == Tournament::FixedPairs;
}

bool TournamentSettings::isSinglePlayerMode() const
{
    return m_tournamentMode == Tournament::SinglePlayers;
}

void TournamentSettings::initializeTournamentOpen()
{
    m_tournamentOpen = true;
    m_originalTournamentOpen = true;
}

void TournamentSettings::setTournamentOpen(bool state)
{
    m_tournamentOpen = state;
}

bool TournamentSettings::tournamentOpen() const
{
    return m_tournamentOpen;
}

bool TournamentSettings::tournamentOpenChanged() const
{
    return m_tournamentOpen != m_originalTournamentOpen;
}

void TournamentSettings::setBoogerScore(int score)
{
    m_boogerScore = score;
}

const int &TournamentSettings::boogerScore() const
{
    return m_boogerScore;
}

void TournamentSettings::setBoogersPerRound(int count)
{
    m_boogersPerRound = count;
}

int TournamentSettings::boogersPerRound() const
{
    return m_boogersPerRound;
}

void TournamentSettings::setNoTableNumbers(bool state)
{
    m_noTableNumbers = state;
}

bool TournamentSettings::noTableNumbers() const
{
    return m_noTableNumbers;
}

void TournamentSettings::setAutoSelectPairs(bool state)
{
    m_autoSelectPairs = state;
}

bool TournamentSettings::autoSelectPairs() const
{
    return m_autoSelectPairs;
}

void TournamentSettings::setSelectByLastRound(bool state)
{
    m_selectByLastRound = state;
}

bool TournamentSettings::selectByLastRound() const
{
    return m_selectByLastRound;
}

void TournamentSettings::setIncludeOpponentGoals(bool state)
{
    m_includeOpponentGoals = state;
}

bool TournamentSettings::includeOpponentGoals() const
{
    return m_includeOpponentGoals;
}

void TournamentSettings::setGrayOutUnimportantGoals(bool state)
{
    m_grayOutUnimportantGoals = state;
}

bool TournamentSettings::grayOutUnimportantGoals() const
{
    return m_grayOutUnimportantGoals;
}

void TournamentSettings::setMarker(Markers::Type type, int id)
{
    m_marker[type] = id;
}

void TournamentSettings::setMarkers(const QHash<Markers::Type, int> &markers)
{
    QHash<Markers::Type, int>::const_iterator it;
    for (it = markers.constBegin(); it != markers.constEnd(); it++) {
        m_marker[it.key()] = it.value();
    }
}

int TournamentSettings::marker(Markers::Type type) const
{
    return m_marker.value(type);
}

QHash<Markers::Type, int> TournamentSettings::specialMarkers() const
{
    QHash<Markers::Type, int> markers;
    for (const auto type : Markers::specialMarkers) {
        markers[type] = m_marker.value(type);
    }
    return markers;
}

void TournamentSettings::setSinglesManagement(const Markers::SinglesManagement &data)
{
    m_singlesManagement = data;
}

const Markers::SinglesManagement &TournamentSettings::singlesManagement() const
{
    return m_singlesManagement;
}

void TournamentSettings::setRegistrationColumns(const QVector<RegistrationColumns::Column> &data)
{
    m_registrationColumns = data;
}

const QVector<RegistrationColumns::Column> &TournamentSettings::registrationColumns() const
{
    return m_registrationColumns;
}

void TournamentSettings::setRegistrationListHeaderHidden(bool state)
{
    m_registrationListHeaderHidden = state;
}

bool TournamentSettings::registrationListHeaderHidden() const
{
    return m_registrationListHeaderHidden;
}

void TournamentSettings::setScoreType(Tournament::ScoreType type)
{
    m_scoreType = type;
}

Tournament::ScoreType TournamentSettings::scoreType() const
{
    return m_scoreType;
}

void TournamentSettings::setDrawSettings(const Draw::Settings &settings)
{
    m_drawSettings = settings;
}

const Draw::Settings &TournamentSettings::drawSettings() const
{
    return m_drawSettings;
}

void TournamentSettings::setScheduleSettings(const Schedule::Settings &settings)
{
    m_scheduleSettings = settings;
}

const Schedule::Settings &TournamentSettings::scheduleSettings() const
{
    return m_scheduleSettings;
}

void TournamentSettings::setBookingSettings(const QString &settings)
{
    m_bookingSettings = settings;
}

const QString &TournamentSettings::bookingSettings() const
{
    return m_bookingSettings;
}

void TournamentSettings::setBookingEnabled(bool state)
{
    m_bookingEnabled = state;
    Q_EMIT bookingEnabledChanged(state);
}

bool TournamentSettings::bookingEnabled() const
{
    return m_bookingEnabled;
}
